import pickle
from dataclasses import asdict
from pathlib import Path
from typing import Any, Callable, Optional

from pydantic import BaseModel, PositiveInt, PrivateAttr

from openiziai.schemas import GPTModel, Message

from .contexts import Context

Exporter = Callable[[list[Message]], Any]


class ContextHandler(BaseModel):
    """Implementa os métodos necessários para criar o contexto do Agente."""
    max_context_length: PositiveInt
    context_store: Path
    agent_model: Optional[GPTModel] = None
    _history: list[Message] = PrivateAttr(default=[])
    _context: Context = PrivateAttr(default=None)

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        self._context = Context(
            agent_model=self.agent_model,  # pyright: ignore
            history=[],
            max_context_length=self.max_context_length,
            context_store=self.context_store,
        )

    def create(self, overwrite: bool = False):
        """Cria ou limpa o contexto de um agente."""
        if len(self._history) > 0 and not overwrite:
            raise Exception(
                (
                    'Já existe um histórico para esse contexto.'
                    'Caso queira sobescrever defina `overwrite=True`.'
                )
            )

        self._history = []

    @property
    def history(self) -> list[Message]:
        """Retorna o histórico de interações do Agente."""
        return self._history[-self.max_context_length :]

    def add(self, message: Message):
        """Adiciona uma interação no contexto."""
        self._history.append(message)
        self._context.history = self._history

    def save(self):
        """Salva o contexto como um pickle."""
        self._context.history = self._history
        self.context_store.mkdir(parents=True, exist_ok=True)
        with open(
            self.context_store / f'context_{self._context.id}.pkl', 'wb'
        ) as file:
            pickle.dump(asdict(self._context), file, protocol=-1)

    def export_history(self, func: Exporter):
        """Executa uma função exporter para salvar todo o histório de
        interações do Agente.
        """
        func(self._history)

    @property
    def context(self) -> Context:
        """Retorna o contexto do Agente."""
        return self._context



"""Ferramentas gerais."""

import functools
import random

from trio import sleep


def exponential_backoff(retries: int = 64, base_delay: float = 1):
    """Implementa o método de exponential backoff para retries em funções
    assíncronas.
    """

    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            attempt = 0
            while attempt < retries:
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if attempt == retries - 1:
                        raise e
                    delay = base_delay * (2 ** (attempt)) + random.uniform(
                        0, 1
                    )
                    attempt += 1
                    await sleep(delay)

        return wrapper

    return decorator



import pickle
from datetime import datetime
from functools import lru_cache
from pathlib import Path

from openiziai.schemas import DataDict, Pipeline


@lru_cache()
def prep_data(pipeline: Pipeline, persist: bool = True) -> DataDict:
    """Aplica um pipeline e retorna os dados no formato DataDict.

    Args:
        pipeline (Pipeline): Pipeline a ser executado. Deve contem o método
            run() implementado.
        persist (bool): Se o dado criado deve ser persistido em um pickle.
            Padrão True.

    Returns:
        DataDict: Dados em um dicionário que contém a chave `data`.
    """
    _pipelined_data = pipeline.run()

    if persist:
        path = Path(__file__).cwd() / 'data' / 'pipelined_data'
        path.mkdir(parents=True, exist_ok=True)
        file = path / f'pipelined_data_{datetime.now().strftime("%Y%m%d")}.pkl'

        with open(file, 'wb') as f:
            pickle.dump(_pipelined_data, f, protocol=-1)

    return {'data': _pipelined_data}



from pathlib import Path
from uuid import uuid4

from pydantic import Field, PositiveInt
from pydantic.dataclasses import dataclass

from openiziai.schemas import GPTModel, Message


@dataclass
class Context:
    """Contexto de um Agente."""
    agent_model: GPTModel
    history: list[Message]
    max_context_length: PositiveInt
    context_store: Path
    id: str = Field(init=False, default=f'context-{uuid4()}')



"""Define o objeto de Task."""

from pydantic.dataclasses import dataclass


@dataclass
class Task:
    """Define a task que será executada por um modelo ou agente."""

    backstory: str
    short_backstory: str
    role: str
    goal: str



"""Módulo que define o principal Agente."""

from typing import Any, Optional, Self

from openai import OpenAI
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic.dataclasses import dataclass

from openiziai.schemas import GPTModel, Message
from openiziai.task import Task


@dataclass
class PromptResponse:
    """Informações do prompt construído."""

    id: str
    prompt: str
    response: str | None
    temperature: float
    total_tokens: int | None
    fine_tuned_model: str


class Agent(BaseModel):
    """Classe que constrói o Agente especializado utilizando um modelo GPT."""

    client: OpenAI = Field(description='Client da OpenAI.')
    model: Optional[GPTModel] = Field(
        default=None, description='A entidade do modelo fine tuned.'
    )
    fine_tuned_model: Optional[str] = Field(
        default=None,
        description='Nome do modelo que será utilizado no Agente.',
    )
    task: Optional[Task] = Field(
        default=None, description='Task a ser executada pelo Agente.'
    )
    _template: str
    _full_context: list[dict[str, Any]]
    _context: list[dict[str, Any]]
    _fine_tuned_model: str

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **data: Any) -> None:
        """Cria um novo agente analisando e validando o input.

        Parameters:
            client (OpenAI): Client da OpenAI.
            model (GPTModel): A entidade do modelo fine tuned.
            fine_tuned_model (str): Modelo que será utilizado no Agente.
            task (Task): Task a ser executada pelo Agente.

        Examples:
            >>> from openai import OpenAI
            >>>
            >>>
            >>> client = OpenAI()
            >>> task = Task(
            ...    backstory='backstory',
            ...    short_backstory='short_backstory',
            ...    role='role',
            ...    goal='goal',
            ... )
            >>>
            >>> my_model = GPTModel(
            ...    name='my_model',
            ...    task=task,
            ...    base_model='gpt-3.5-turbo',
            ... )
            >>>
            >>> Agent(client=client, model=my_model)
            >>> # OU
            >>> Agent(client=client, fine_tuned_model='my_model', task=task)
        """
        super().__init__(**data)
        self._template = self._build_template()
        self._fine_tuned_model = (
            self.model.name if self.model else self.fine_tuned_model
        )  # pyright: ignore

    @model_validator(mode='after')
    def validate_model_info(self) -> Self:
        """Valida se o `model` ou se `fine_tuned_model` e `task` foram
        declarados.
        """
        if not self.model:
            if not self.fine_tuned_model and not self.task:
                raise ValueError(
                    'Precisa de um `model` ou um `fine_tuned_model` e `task`.'
                )
            self.model = GPTModel(
                name=self.fine_tuned_model,  # pyright: ignore
                task=self.task,  # pyright: ignore
            )

        return self

    def _build_template(self) -> str:
        backstory = (
            self.model.task.short_backstory
            if self.model
            else self.task.short_backstory  # pyright: ignore
        )
        role = self.model.task.role if self.model else self.task.role  # pyright: ignore
        goal = self.model.task.goal if self.model else self.task.goal  # pyright: ignore

        template = f"""{backstory}.
        {{'your_role': {role}, 'your_goal': {goal}}}
        """

        return template

    def prompt(
        self,
        prompt: Optional[str] = None,
        *,
        history: Optional[list[Message]] = None,
        temperature: float = 0.5,
        max_tokens: int = 1000,
    ) -> PromptResponse:
        """Executa o prompt para o modelo de fine tuning.

        Args:
            prompt (str): Prompt.
            history (list[Message] | None): Histórico de interações do Agente.
            temperature (float): Temperatura que controla a criatividade ao
                construir a resposta
            max_tokens (int): Máximo de tokens que deve conter nas respostas.
                Valores maiores trarão respostas maiores porém terá maior
                custo.

        Returns:
            PromptResponse: Informações do prompt construído.
        """
        if history:
            _history = [m.model_dump() for m in history]
            messages = [{'role': 'system', 'content': self._template}].extend(
                _history
            )
            _prompt = _history[-1].get('content')
        elif prompt:
            _prompt = prompt
            messages = [
                {
                    'role': 'system',
                    'content': self._template,
                },
                {
                    'role': 'user',
                    'content': _prompt,
                },
            ]
        else:
            raise ValueError('`prompt` ou `history` precisam ser passados.')

        result = self.client.chat.completions.create(
            messages=messages,  # pyright: ignore
            model=self._fine_tuned_model,
            temperature=temperature,
            max_tokens=max_tokens,
        )

        response = PromptResponse(
            id=result.id,
            prompt=_prompt,  # pyright: ignore
            response=result.choices[0].message.content,
            temperature=temperature,
            total_tokens=result.usage.total_tokens if result.usage else None,
            fine_tuned_model=self._fine_tuned_model,
        )

        return response



import json
import random
from datetime import datetime
from math import ceil
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4

import trio
from openai import OpenAI
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
)

from openiziai.schemas import DataDict
from openiziai.task import Task
from openiziai.utils import exponential_backoff


class TrainDataTool(BaseModel):
    """Cria dados preparados para fine tuning.

    Cria os dados para construir um fine tuning de um modelo LLM centrado
    em uma task contendo: backstory, role e o goal.
    """

    client: OpenAI = Field(default=None, description='Client da OpenAI.')
    data: DataDict = Field(
        description='Dados utilizados para construir o dado de treino.',
    )
    task: Task = Field(
        description='Descrição da task que o modelo deverá executar.',
    )
    model: str = Field(
        default='gpt-3.5-turbo-125',
        description='Modelo GPT que criará o dado de fine tuning.',
    )

    _n_examples: int = PrivateAttr(default=None)
    _n_batch: int = PrivateAttr(default=None)
    _root: Path = PrivateAttr(default_factory=Path.cwd)
    _train_data_dir: Path
    _template: str
    _hash: str = PrivateAttr(default_factory=uuid4)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **data: Any) -> None:
        """Cria um novo modelo analisando e validando o input.

        Args:
            client (OpenAI): Client da OpenAI.
            data (DataDict): Dados utilizados para construir o dado de treino.
            task (Task): Descrição da task que o modelo treinado irá executar.
            model (str): Modelo GPT usado para criar os dados de treino.
                Padrão gpt-3.5-turbo-125.
        """
        super().__init__(**data)
        self._template = """You are generating data which will be used to train a machine learning model.
        To describe how the model will be trained, you will be given a high-level description in a dict with:
        ```{{'backstory': $backstory_goes_here, 'role': $role_goes_here, 'goal': $goal_goes_here, 'data': $data_goes_here}}
        From that, you will generate data samples, each with a prompt/response pair. You must return the sample in jsonl format with the exact informations:
        ```{{'prompt': $prompt_goes_here, 'response': $response_goes_here}}```
        Only one prompt/response pair should be generated by turn. For each turn, make the example slightly more complex than the last, while ensuring diversity.
        Here is the high-level description to generate the samples:
        {description}
        """  # noqa

        self._train_data_dir = self._create_dir_if_not_exist(
            self._root / 'data' / 'train'
        )

    @property
    def id(self) -> str:
        """Hash da classe."""
        return self._hash

    @property
    def train_data_dir(self) -> str:
        """Diretório que os dados de treino serão salvos."""
        return str(self._train_data_dir)

    @staticmethod
    def _create_dir_if_not_exist(dir: Path) -> Path:
        dir.mkdir(parents=True, exist_ok=True)

        return dir

    @exponential_backoff()
    async def create_examples(  # noqa
        self,
        n_examples: int,
        temperature: float,
        max_tokens: int,
        max_context_length: int,
        sender: trio.abc.SendChannel[list[dict[str, str]]],
    ) -> None:
        """Cria exemplos com o par: prompt/response.

        Utilizando um modelo GPT, cria exemplos com o par prompt/response no
            formato de json.

        Args:
            n_examples (int): Número de exemplos que serão criados.
            temperature (float): Temperatura que controla a criatividade ao
                criar os exemplos.
            max_tokens (int): Máximo de tokens que deve conter nas respostas.
                Valores maiores trarão prompts e responses maiores porém
                terá maior custo.
            max_context_length (int): Quantidade de exemplos que devem ser
                usados como contexto ao criar o próximo.
            sender (SendChannel): Canal de Pub par aplicar o Pub/Sub.
        """
        description = dict(
            backstory=self.task.backstory,
            role=self.task.role,
            goal=self.task.goal,
            **self.data,
        )
        prompt = self._template.format(description=description)

        short_system = {'role': 'system', 'content': self.task.short_backstory}

        examples = []

        for _ in range(n_examples):
            context = (
                [
                    {'role': 'assistant', 'content': example}
                    for example in examples
                ]
                if len(examples) <= max_context_length
                else [
                    {'role': 'assistant', 'content': example}
                    for example in random.sample(examples, 8)
                ]
            )

            messages = [
                {
                    'role': 'system',
                    'content': prompt,
                }
            ] + context

            result = self.client.chat.completions.create(
                model=self.model,
                messages=messages,  # pyright: ignore
                temperature=temperature,
                max_tokens=max_tokens,
            )

            _response = result.choices[0].message.content
            if not _response:
                print('Algo de errado aconteceu ao criar o prompt.')
                continue
            response = json.loads(_response)
            example = {
                'messages': [
                    short_system,
                    {'role': 'user', 'content': response.get('prompt')},
                    {
                        'role': 'assistant',
                        'content': response.get('response'),
                    },
                ]
            }

            examples.append(example)
        async with sender:
            await sender.send(examples)

    async def create_train_file(self, receiver: trio.abc.ReceiveChannel):
        """Salva os exemplos gerados em um arquivo jsonl.

        Método Sub do sistema Pub/Sub.

        Args:
            receiver (ReceiveChannel): Canal de receive do sistema de Pub/Sub.
        """
        self._file = (
            self._train_data_dir
            / f'train_{self.id}_{datetime.now().strftime("%Y%m%d")}.jsonl'
        )

        if self._file.exists():
            self._file.unlink()

        async with receiver:
            async for results in receiver:
                async with await trio.open_file(
                    self._file, 'a', encoding='utf-8'
                ) as file:
                    for result in results:
                        await file.write(json.dumps(result) + '\n')

    async def create_train_data(  # noqa
        self,
        n_examples: int,
        n_batch: int,
        temperature: float = 0.5,
        max_tokens: int = 1000,
        max_context_length: int = 8,
    ) -> str:
        """Cria os dados de treino.

        Aplica o sistema de Pub/Sub para criar os exemplos e salvar em um jsonl
            a medida que os batches de treino ficam prontos.

        Args:
            n_examples (int): Número de exemplos que devem ser criados.
            n_batch (int): Número de batches para serem executados de forma
                concorrente.
            temperature (float): Nível de criatividade do modelo ao criar os
                exemplos. Padrão 0.5.
            max_tokens (int): Máximo de tokens que deve conter nas respostas.
                Valores maiores trarão prompts e responses maiores porém
                terá maior custo.
            max_context_length (int): Quantidade de exemplos que devem ser
                usados como contexto ao criar o próximo.

        Returns:
            str: Nome do arquivo JSONL criado.
        """
        sender, receiver = trio.open_memory_channel(0)
        self._n_batch = n_batch
        _n_examples = ceil(n_examples / self._n_batch)
        self._n_examples = _n_examples * self._n_batch

        async with trio.open_nursery() as nursery:
            async with sender, receiver:
                for _ in range(self._n_batch):
                    nursery.start_soon(
                        self.create_examples,
                        _n_examples,
                        temperature,
                        max_tokens,
                        max_context_length,
                        sender.clone(),
                    )
                    nursery.start_soon(
                        self.create_train_file,
                        receiver.clone(),
                    )

        return str(self._file)

    @property
    def n_examples(self) -> Optional[int]:
        """Número de exemplos criados."""
        if not self._n_examples:
            print(
                'Você deve criar um dataset de treino com `create_train_data`.'
            )
            return None
        return self._n_examples

    @property
    def n_batch(self) -> Optional[int]:
        """Número de batches utilizados."""
        if not getattr(self, '_n_batch'):
            print(
                'Você deve criar um dataset de treino com `create_train_data`.'
            )
            return None
        return self._n_batch

    def execute(  # noqa
        self,
        n_examples: int,
        n_batch: int,
        temperature: float = 0.5,
        max_tokens: int = 1000,
        max_context_length: int = 8,
    ) -> str:
        """Executa todo o pipeline para criar os dados de treino.
        Args:
            n_examples (int): Número de exemplos que devem ser criados.
            n_batch (int): Número de batches para serem executados de forma
                concorrente.
            temperature (float): Nível de criatividade do modelo ao criar os
                exemplos. Padrão 0.5.
            max_tokens (int): Máximo de tokens que deve conter nas respostas.
                Valores maiores trarão prompts e responses maiores porém
                terá maior custo.
            max_context_length (int): Quantidade de exemplos que devem ser
                usados como contexto ao criar o próximo.

        Returns:
            str: Nome do arquivo JSONL criado.
        """  # noqa
        file = trio.run(
            self.create_train_data,
            n_examples,
            n_batch,
            temperature,
            max_tokens,
            max_context_length,
        )

        return file

    @property
    def file(self) -> Optional[str]:
        """Nome do arquivo com os dados de treino."""
        if not hasattr(self, '_file'):
            print('Nenhum dado de treino foi criado com esta instância.')
            return None
        return str(self._file)

    def __repr__(self) -> str:
        return (
            'TrainDataTool('
            f'task={self.task}, '
            f'n_examples={self.n_examples}, '
            f'n_batch={self.n_batch}, '
            f'file={self.file}'
            ')'
        )



from pathlib import Path
from typing import Optional

from pydantic import BaseModel, PositiveInt, PrivateAttr

from openiziai.contexts import Context, ContextHandler, Exporter
from openiziai.schemas import Message

from .agent import Agent


class AgentManager(BaseModel):
    """Gerenciador de contexto capaz de gerenciar o contexto de interações do
    agente.
    """

    agent: Agent
    pre_context: Optional[Context] = None
    context_store: Path = Path().cwd() / 'data/contexts'
    max_context_length: PositiveInt = 10
    context_exporter: Optional[Exporter] = None
    _ctx_handler: ContextHandler = PrivateAttr(default=None)

    def __enter__(self) -> 'AgentManager':
        if not self.pre_context:
            self._ctx_handler = ContextHandler(
                agent_model=self.agent.model,
                max_context_length=self.max_context_length,
                context_store=self.context_store,
            )
            self._ctx_handler.create()
        else:
            self._ctx_handler = ContextHandler(
                max_context_length=self.pre_context.max_context_length,
                context_store=self.pre_context.context_store,
                agent_model=self.pre_context.agent_model,
            )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False

        self._ctx_handler.save()
        if self.context_exporter:
            self._ctx_handler.export_history(self.context_exporter)

    def prompt(
        self,
        prompt: str,
        *,
        temperature: float = 0.5,
        max_tokens: int = 1000,
    ):
        """Executa o prompt para o Agente.

        Args:
            prompt (str): Prompt.
            temperature (float): Temperatura que controla a criatividade ao
                construir a resposta
            max_tokens (int): Máximo de tokens que deve conter nas respostas.
                Valores maiores trarão respostas maiores porém terá maior
                custo.

        Returns:
            PromptResponse: Informações do prompt construído.
        """

        self._ctx_handler.add(Message(content=prompt))
        response = self.agent.prompt(
            history=self._ctx_handler.history,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        if response.response:
            self._ctx_handler.add(
                Message(role='assistant', content=response.response)
            )

        return response

    @property
    def context(self) -> Context:
        """O contexto completo do Agente."""
        return self._ctx_handler.context



"""Define os schemas utilizados."""

import sys
from datetime import datetime
from typing import Any, Literal, Optional, Protocol

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

from openiziai.task import Task

if sys.version_info < (3, 12):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class DataDict(TypedDict):
    """Schema dos dados que serão utilizados para criar os treinos."""

    data: Any


class Pipeline(Protocol):
    """Define o método que deve ser implementado em um pipeline."""

    def run(self): ...


@dataclass
class GPTModel:
    """Entidade do modelo GPT treinado."""

    name: str
    task: Task
    base_model: Optional[str] = None
    created_at: datetime = Field(init=False, default_factory=datetime.now)


class Message(BaseModel):
    """Representa a interação de um Agente."""
    role: Literal['assistant', 'user'] = 'user'
    content: str



"""Disponibiliza o fine tuning do modelo."""

from enum import Enum
from pathlib import Path
from typing import Any, Optional

from openai import OpenAI
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    field_validator,
)

from openiziai.schemas import GPTModel
from openiziai.task import Task


class JobStatus(Enum):
    """Status do job de fine tuning."""

    VALIDATING = 'validating_files'
    QUEUED = 'queued'
    COMPLETED = 'succeeded'
    RUNNING = 'running'
    FAILED = 'failed'
    CANCELLED = 'cancelled'


class FineTuning(BaseModel):
    """Classe que gerencia os métodos necessários para construir um modelo de
    fine tuning.
    """

    client: OpenAI = Field(description='Client da OpenAI.')
    train_file: Path | str = Field(
        description='Caminho até o arquivo de treino.'
    )
    task: Task = Field(
        description='Task em que o modelo deve se especializar.'
    )
    base_model: str = Field(
        default='gpt-3.5-turbo', description='Modelo base que será refinado.'
    )
    _file_id: str = PrivateAttr(default=None)
    _job_id: str = PrivateAttr(default=None)
    _job_status: JobStatus = PrivateAttr(default=None)
    _model: GPTModel = PrivateAttr(default=None)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **data: Any) -> None:
        """Cria um novo fine tuning analisando e validando o input.

        Args:
            client (OpenAI): Client da OpenAI.
            train_file (Path|str): Caminho até o arquivo de treino.
            task (Task): Task em que o modelo deve se especializar.
            base_model (str): Modelo base que será refinado.
        """
        super().__init__(**data)

    @field_validator('train_file')
    @classmethod
    def size_must_be_minor_512(cls, v: Path | str) -> Path | str:
        """Valida se o arquivo indicado possui tamanho menor que 512MB."""
        MAX_FILE_SIZE = 512000000
        _v = Path(v) if isinstance(v, str) else v
        size = _v.stat().st_size
        if size > MAX_FILE_SIZE:
            raise ValueError(
                'max_file_size_error',
                f'Arquivo de treino deve conter no máximo 512MB. {size / 100000}MB',  # noqa
            )

        return v

    def upload_file_to_openai(self) -> 'FineTuning':
        """Envia o arquivo de treino para a plataforma da OpenAI."""
        self._file_id = self.client.files.create(
            file=open(self.train_file, 'rb'), purpose='fine-tune'
        ).id

        return self

    @property
    def file_id(self) -> Optional[str]:
        """ID do arquivo de treino na plataforma OpenAI."""
        if not self._file_id:
            print('Nenhum dado foi enviado ainda.')
            return None
        return self._file_id

    def start(self) -> Optional['FineTuning']:
        """Inicia o job de fine tuning."""
        file_id = self.file_id
        if not file_id:
            return None

        job = self.client.fine_tuning.jobs.create(
            training_file=file_id, model=self.base_model
        )
        self._job_id = job.id
        print(
            f'Fine tuning started: {self._job_id}.',
            'Veja o status com `.status`.',
        )
        return self

    @property
    def job_id(self) -> Optional[str]:
        """ID do job de fine tuning na plataforma OpenAI."""
        if not getattr(self, '_job_id'):
            print('Nenhum fine tuning foi iniciado.')
            return None
        return self._job_id

    @property
    def status(self) -> Optional[str]:
        """Status do job de fine tuning."""
        job_id = self._job_id
        if not job_id:
            print('Nenhum fine tuning foi iniciado.')
            return None

        if self._job_status != JobStatus.COMPLETED:
            _job_status = self.client.fine_tuning.jobs.retrieve(job_id).status
            self._job_status = JobStatus(_job_status)

        return self._job_status.name

    @property
    def model(self) -> Optional[GPTModel]:
        """Modelo criado do fine tuning."""
        if self._model:
            return self._model

        job_id = self.job_id
        if not job_id:
            print('Nenhum fine tuning foi iniciado.')
            return None

        model_name = self.client.fine_tuning.jobs.retrieve(
            job_id
        ).fine_tuned_model
        if not model_name:
            print(f'Modelo não disponível. Status: {self.status}')
            return None

        self._model = GPTModel(
            name=model_name, base_model=self.base_model, task=self.task
        )

        return self._model

    def __repr__(self) -> str:
        return (
            'FineTuning('
            f'client={self.client}, '
            f'train_file={self.train_file}, '
            f'task={self.task}, '
            f'base_model="{self.base_model}", '
            f'model={self.model}'
            ')'
        )
